# Script to fill our Community Dashboard spreadsheet from Launchpad data.

from launchpadlib.launchpad import Launchpad
from launchpadlib.credentials import Credentials
import threading
import getpass
import string
import gdata.spreadsheet.service
import gdata.service
import gdata.spreadsheet
import os
import re
import time
import getopt
import sys
from datetime import datetime
from optparse import OptionParser
import sys

LP_SERVICE = 'production'

TITLE_HEADER = 'date' # the header name we used as title for spreadsheet rows

# List of bug status:
#   [ 'New', 'Incomplete', 'Invalid', 'Opinion',
#     'Confirmed', 'Queued',
#     'Superseded', 'Merged', 'Approved', 'Rejected' ]
OPEN_BUGS_STATUS = ['Confirmed', 'Triaged', 'In Progress']
NEW_BUGS_STATUS = OPEN_BUGS_STATUS + ['New', 'Fix Committed']

# List of merge status:
#   [ 'Work in progress', 'Needs review', 'Code failed to merge', 'Queued',
#     'Superseded', 'Merged', 'Approved', 'Rejected' ]
NEW_MERGE_STATUS = ['Needs review','Code failed to merge', 'Approved']


STATUS_FILE = os.path.expanduser("~/.community_dashboard.status")

def unqualified_bug_filter_spec(name, project):
    return {
        'name': name,
        'filters': {
            'project': project,
            'status': NEW_BUGS_STATUS,
            'assignee': None,
        }
    }

def assigned_bug_filter_spec(name, team):
    return {
        'name': name,
        'filters': {
            'status': OPEN_BUGS_STATUS,
            'assignee': team,
            'importance': ['Critical', 'High', 'Medium', 'Low', 'Undecided'], #no Wishlists!
        }
    }

def unqualified_project_merge_filter_spec(name, project):
    return {
        'name': name,
        'filters': {
            'project': project,
            'status': NEW_MERGE_STATUS,
        }
    }

def unqualified_branch_merge_filter_spec(name, branch):
    return {
        'name': name,
        'filters': {
            'branch': branch,
            'status': NEW_MERGE_STATUS,
        }
    }

# New bugs stats
server_newbugs = unqualified_bug_filter_spec('server-newbugs','openobject-server')
addons_newbugs = unqualified_bug_filter_spec('addons-newbugs','openobject-addons')
client_newbugs = unqualified_bug_filter_spec('client-newbugs','openobject-client')
web_newbugs = unqualified_bug_filter_spec('web-newbugs','openerp-web')

# Assigned bugs stats
framework_bugs = assigned_bug_filter_spec('framework-assigned','openerp-dev-framework')
client_bugs = assigned_bug_filter_spec('client-assigned','openerp-dev-gtk')
web_bugs = assigned_bug_filter_spec('web-assigned','openerp-dev-web')
addons1_bugs = assigned_bug_filter_spec('addons1-assigned','openerp-dev-addons1')
addons2_bugs = assigned_bug_filter_spec('addons2-assigned','openerp-dev-addons2')
addons3_bugs = assigned_bug_filter_spec('addons3-assigned','openerp-dev-addons3')

#Active merge reviews - trunk
server_merge_trunk = unqualified_branch_merge_filter_spec('server-merge','~openerp/openobject-server/trunk')
client_merge_trunk = unqualified_branch_merge_filter_spec('client-merge','~openerp/openobject-client/trunk')
addons_merge_trunk = unqualified_branch_merge_filter_spec('addons-merge','~openerp/openobject-addons/trunk')
web_merge_trunk = unqualified_branch_merge_filter_spec('web-merge','~openerp/openerp-web/trunk')

#Active merge reviews - 6.0
server_merge_60 = unqualified_branch_merge_filter_spec('server60-merge','~openerp/openobject-server/6.0')
client_merge_60 = unqualified_branch_merge_filter_spec('client60-merge','~openerp/openobject-client/6.0')
addons_merge_60 = unqualified_branch_merge_filter_spec('addons60-merge','~openerp/openobject-addons/6.0')
web_merge_60 = unqualified_branch_merge_filter_spec('web60-merge','~openerp/openobject-client-web/6.0')

#Active merge reviews - 6.1
server_merge_61 = unqualified_branch_merge_filter_spec('server61-merge','~openerp/openobject-server/6.1')
addons_merge_61 = unqualified_branch_merge_filter_spec('addons61-merge','~openerp/openobject-addons/6.1')
web_merge_61 = unqualified_branch_merge_filter_spec('web61-merge','~openerp/openerp-web/6.1')

#Active merge reviews - 7.0
server_merge_70 = unqualified_branch_merge_filter_spec('server70-merge','~openerp/openobject-server/7.0')
addons_merge_70 = unqualified_branch_merge_filter_spec('addons70-merge','~openerp/openobject-addons/7.0')
web_merge_70 = unqualified_branch_merge_filter_spec('web70-merge','~openerp/openerp-web/7.0')

# verbatim from gdata api doc
def StringToDictionary(row_data):
  result = {}
  for param in row_data.split():
    name, value = param.split('=')
    result[name] = value
  return result

class lpServer(threading.Thread):
    host = 'localhost'
    port = 8200
    cachedir = os.path.expanduser("~/.launchpadlib/credentials")
    launchpad = False

    def __init__(self, email, pwd, ss_key, ws_key):
        self.gd_client = gdata.spreadsheet.service.SpreadsheetsService()
        self.gd_client.email = email
        self.gd_client.password = pwd
        self.ss_key = ss_key
        self.ws_key = ws_key
        self.gd_client.source = 'OpenERP Community Dashboard'
        self.gd_client.ProgrammaticLogin()
        super(lpServer, self).__init__()
        self.launchpad = self.get_lp()

    def _get_spreadsheet(self):
        feed = self.gd_client.GetSpreadsheetsFeed()
        spreadsheet_key = feed.entry[0].id.text.rsplit('/',1)[1]
        return spreadsheet_key

    def _get_worksheet(self,ss_key):
        new_sheet_name = False
        sheet_list = []
        work_feed = self.gd_client.GetWorksheetsFeed(ss_key)
        link = work_feed.GetAlternateLink().href
        print "Related Spread Sheet link:::::::::",link
        current_date = datetime.now().strftime('%Y-%m')
        for sheet in work_feed.entry:
            if current_date == sheet.title.text:
                return sheet.id.text.rsplit('/',1)[1]
        current_sheet = self.gd_client.AddWorksheet(title=current_date, row_count=1, col_count=100, key=ss_key)
        return current_sheet.id.text.rsplit('/',1)[1]

    def _clean_name(self, name):
        return re.sub(r'[^a-zA-Z0-9-_]', '', name).lower()

    def _update_data_row(self, ss_key, ws_key, value_dict):
        # check if a row with this title already exists
        row_title = value_dict[TITLE_HEADER]
        query = gdata.spreadsheet.service.ListQuery()
        query['sq'] = '%s=%s' % (TITLE_HEADER, row_title)
        feed = self.gd_client.GetListFeed(ss_key, ws_key, query=query)
        if len(feed.entry):
            # row exists, we can simply update it!
            print feed.entry[0]
            # entry.custom is a dict of the row data
            # and we need to preserve the current data
            current_data_custom = feed.entry[0].custom
            row_data = {}
            for header in current_data_custom:
                row_data[header] = current_data_custom[header].text
            row_data.update(value_dict)
            entry = self.gd_client.UpdateRow(feed.entry[0], row_data)
            #print 'Updated row', entry
        else:
            # row doesnt exist, we need to insert it!
            entry = self.gd_client.InsertRow(value_dict, ss_key, ws_key)
            #print 'Inserted entry', entry

    def _update_header_row(self, ss_key, ws_key, headers):
        # check existing headers to see if any header is missing, and add them
        query = gdata.spreadsheet.service.CellQuery()
        query['min-row'] = '1'
        query['max-row'] = '1'
        feed = self.gd_client.GetCellsFeed(ss_key, ws_key, query=query)
        current_headers = []
        max_header_col_index = 0
        # need to find real header column so we know where to insert the new ones
        for cell in feed.entry:
            current_headers.append(cell.content.text)
            max_header_col_index = max(int(cell.cell.col), max_header_col_index)

        # TODO: add check based on worksheet feed metadata to make sure
        #       we never try to update more cells than available

        missing_headers = set(headers).difference(current_headers)

        if TITLE_HEADER in missing_headers:
            # always add the title header first, if present
            self.gd_client.UpdateCell(row=1, col=max_header_col_index+1,
               inputValue=TITLE_HEADER, key=ss_key, wksht_id=ws_key)
            missing_headers.remove(TITLE_HEADER)
            max_header_col_index += 1
        for i, header in enumerate(missing_headers):
            self.gd_client.UpdateCell(row=1, col=max_header_col_index+i+1,
               inputValue=header, key=ss_key, wksht_id=ws_key)

    def insert_row_data(self, ss_key, ws_key, value_dict):
        # first cleanup headers:
        cleaned_dict = {}
        for k, v in value_dict.iteritems():
            cleaned_dict[self._clean_name(k)] = v

        self._update_header_row(ss_key, ws_key, cleaned_dict.keys())
        self._update_data_row(ss_key, ws_key, cleaned_dict)

    def get_lp(self):
        # See http://www.mail-archive.com/launchpad-users@lists.launchpad.net/msg01653.html
        return Launchpad.login_with('openerp-community-dashboard', LP_SERVICE, credentials_file=self.cachedir)

    def get_lp_bug_stat(self, specification):
        stat_name = specification['name']
        filters = specification['filters']
        assert 'project' in filters or 'assignee' in filters, 'Project or Assignee is required in bug filters!'

        # Use Project or Team as source for calling searchTasks()
        if filters.get('project'):
            lp_source = self.launchpad.projects[filters['project']]
            print "Matching LP bugs on project ", filters['project']
            del filters['project']
        elif filters.get('assignee'):
            lp_source = self.launchpad.people[filters['assignee']]
            print "Matching LP bugs on team ", filters['assignee']
            del filters['assignee']

        print "Searching macthing LP bugs with filters ", filters
        matching_bugs = lp_source.searchTasks(**filters)

        # Warning: Launchpad's collections do not support len(), so unfortunately
        #          we need to walk the results to count them!
        # Also, if 'assignee=None' was a criterion, we will need to filter out the
        # results because passing assignee=None to searchTask() is same as not
        # filtering on assignee!

        if 'assignee' in filters and filters['assignee'] is None:
            matching_bugs = [bug for bug in matching_bugs if bug.assignee is None]
        else:
            matching_bugs = [bug for bug in matching_bugs]

        for bug in matching_bugs:
            print bug, bug.importance, bug.status
        print "Found %s matching bugs for %s" % (len(matching_bugs), stat_name)

        return {
            TITLE_HEADER: time.strftime('%m/%d/%Y'),
            stat_name: str(len(matching_bugs))
        }

    def get_lp_merge_stat(self, specification):
        stat_name = specification['name']
        filters = specification['filters']
        assert 'project' in filters or 'branch' in filters, 'Project/Branch required in bug filters!'

        if filters.get('project'):
            lp_source = self.launchpad.projects[filters['project']]
            print "Matching LP merge proposals on project ", filters['project']
            del filters['project']
        elif filters.get('branch'):
            lp_source = self.launchpad.branches.getByUniqueName(unique_name=filters['branch'])
            print "Matching LP merge proposals on branch ", filters['branch']
            del filters['branch']


        # Warning: Launchpad's collections do not support len() so unfortunately 
        #          we need to walk the results to count them!

        matching_merges = [x for x in lp_source.getMergeProposals(**filters)]
        return {
            TITLE_HEADER: time.strftime('%m/%d/%Y'),
            stat_name: str(len(matching_merges))
        }

    def run(self):
        if self.ss_key:
            ss_key = self.ss_key
        else:
            ss_key = self._get_spreadsheet()

        if self.ws_key:
            ws_key = self.ws_key
        else:
            ws_key = self._get_worksheet(ss_key)

        # new and assigned bugs
        for stat in (server_newbugs, web_newbugs, addons_newbugs,
                     framework_bugs, web_bugs, addons1_bugs, addons2_bugs, addons3_bugs):
            self.insert_row_data(ss_key, ws_key, self.get_lp_bug_stat(stat))

        # Getting active review branches for the addons, web, server
        for merge in (server_merge_trunk, addons_merge_trunk, web_merge_trunk,
                      server_merge_61, addons_merge_61, web_merge_61,
                      server_merge_70, addons_merge_70, web_merge_70):
            self.insert_row_data(ss_key, ws_key, self.get_lp_merge_stat(merge))

        return True


if __name__ == '__main__':
    # FIXED: Give the email and password  from command-line
    email = ''
    pwd = ''
    ss_key = ''
    ws_key = ''

    parser = OptionParser()
    parser.add_option("--email", action="store", dest="email")
    parser.add_option("--pwd", action="store", dest="pwd")
    parser.add_option("--ss_key", action="store", dest="ss_key")
    parser.add_option("--ws_key", action="store", dest="ws_key")

    (opts, args) = parser.parse_args()

    email = opts.email
    pwd = opts.pwd
    ss_key = opts.ss_key
    ws_key = opts.ws_key

    if not email:
        email =  raw_input('Email:')
    if not pwd:
        pwd =  raw_input('Password:')

    try:
        lp_server = lpServer(email, pwd, ss_key, ws_key)
        lp_server.run()
    except Exception:
        print "Failure"
        raise
    else:
        print "Success"
        with open(STATUS_FILE, 'w') as f:
            f.write('Last success at %r' % datetime.now())


