"""
Main data structures for the Runbot.
"""
import copy
import optparse
import os
import re
import signal
import simplejson
import subprocess
import socket
import sys
import threading
import time
import traceback
import pdb

import mako.template

import openerprunbot
from openerprunbot.misc import *
from openerprunbot.jobs.install_all import InstallAllJob
runbot_state_lock = threading.Lock()


# Number of build 'slots' per branch group.
POINTS = 5

# Job states
STATE_ALLOCATED = 'allocated'
STATE_BROKEN = 'broken'
STATE_PULLING = 'pulling'
STATE_RUNNING = 'running'
STATE_TESTING = 'testing'

# This constant matches community_dashboard.py.
NEW_MERGE_STATUS = ['Needs review', 'Code failed to merge', 'Approved']

# Hack to allow more ports than strictly necessary.
ADDITIONAL_NUMBER = 50

#Max Merge Proposal to search
MAX_MP_DEFAULT = 15

PDB_SET_TRACE_PATH = os.path.join( os.path.normpath( os.path.dirname( __file__ ) ), '..', "pdb_set_trace")

#----------------------------------------------------------
# OpenERP RunBot Branch
#----------------------------------------------------------

class RunBotBranch(object):
    def __init__(self, runbot, branch, group=None,
        repo_path=None, project_name=None, trigger_build=False):
        """
        Represent a single Launchpad branch, e.g.
        `~openerp-dev/openobject-server/trunk-foo-bar`.
        """
        self.runbot = runbot
        self.group = group
        self.trigger_build = trigger_build

        # e.g. trunk
        self.name = branch.name
        self.name_underscore = underscore(self.name)
        self.name_dashed = dashes(self.name)
        # e.g. ~openerp/openobject-server/trunk
        self.unique_name = branch.unique_name
        self.unique_name_underscore = underscore(self.unique_name)
        self.unique_name_dashes = dashes(self.unique_name)
        # server, addons, or web (both for openobject-client-web and openerp-web)
        if project_name:
            self.project_name = project_name
        else:
            self.project_name = project_name_from_unique_name(self.unique_name)

        self.lp_date_last_modified = branch.date_last_modified
        self.lp_revision_count = branch.revision_count

        self.local_date_last_modified = 0
        self.local_revision_count = 0
        self.merge_count = 0

        # Repository path <Root>/repo/<branchuniquename>
        if repo_path:
            self.repo_path=repo_path
        else:
            self.repo_path=os.path.join(self.runbot.wd,'repo',self.unique_name_underscore)

        self.committer_name = None
        self.committer_xgram = None
        self.committer_email = None
        self.reason_ok = False
        self.merge_target = False
        self.merge_source = False
        self.merge_web_link = False

    def age(self):
        return time.time()-int(self.local_date_last_modified.strftime("%s"))

    def update_launchpad(self,branch):
        self.lp_date_last_modified=branch.date_last_modified
        self.lp_revision_count=branch.revision_count
        if self.local_revision_count != self.lp_revision_count:
            self.local_date_last_modified = self.lp_date_last_modified
            self.local_revision_count = self.lp_revision_count
            if self.trigger_build:
                name = self.project_name
                self.group.need_run_reason.append(name)
                self.group.set_reason_ok( False, False )
                self.group.set_reason_ok( branch.unique_name, True )

#----------------------------------------------------------
# OpenERP RunBot Grouped Branch
#----------------------------------------------------------

class RunBotGroupedBranch(object):
    """
    A single 'branch group' represents a single branch name for all our
    projects. E.g. trunk is available in server, addons, web, and client-web.

    When the corresponding project doesn't have that branch, trunk is used.
    For instance, a branch group can represent
        `~openerp-dev/openobject-server/trunk-foo-bar`
        `~openerp-dev/openobject-addons/trunk-foo-bar`
    and
        `~openerp/openobject-web-client/6.0`
        `~openerp/openerp-web/trunk`
    will be used to complete the group, building the four branches together.

    Alternatively a 'branch group' can be configured. In that case, all
    branches are manually specified; no attempt is done to complete the group.
    """
    def __init__(self, runbot, team_name, name, version, sticky,
            server_branch=None, client_web_branch=None,
            web_branch=None, addons_branches=None, modules=None, job_type=None):
        self.runbot = runbot
        self.sticky = sticky
        self.job_type = job_type

        # Points are build 'slots'
        self.points = [None for x in xrange(POINTS)]

        self.team_name = team_name
        self.name = name # group name or manually chosen name.
        self.name_underscore = underscore(self.name)
        self.name_dashes = dashes(self.name)
        self.version = version # '6.0', '6.1', '7.0'  or 'trunk'

        # Reason to do a build, e.g. because 'addons' was commited to.
        self.need_run_reason = []

        self.server_branch = server_branch
        self.web_branch = client_web_branch or web_branch
        #self.addons_branches = addons_branches

        #Split with comma into branches name, for use only one input text
        self.addons_branches = []
        for branch in (addons_branches or []):
            if branch:
               self.addons_branches.extend( branch.split(',') )

        self.server = None
        self.web = None
        self.addons = []

        self.json_path=os.path.join(runbot.wd,'static',"%s-%s.json"%(team_name, self.name.replace('_','-').replace('.','-')))

        # List of modules to install instead of `all`.
        self.modules = modules.split(',') if modules else []

        self.configured = False

        # A normal build_number is maxint, a build_number is assigned when a
        # build is manually requested or a triggering branch has a new revision.
        # The build number is used to assign a position in the queue.
        # The value will be reset to maxint after the build is done.
        self.build_number = sys.maxint

        # True when it is not possible to provide default branches based on
        # the name.
        self.wrong_matching = False

    def add_point(self, j):
        assert j.state == STATE_ALLOCATED
        job0 = self.points[0]
        if job0:
            job0.stop()
            try:
                del self.runbot.jobs[job0.job_id]
            except KeyError:
                pass
        self.points = self.points[1:] + [j]

    def complete_point(self, j):
        assert j.state in (STATE_RUNNING, STATE_BROKEN)
        for p in self.points + [None]:
            if p and p.job_id == j.job_id:
                break
        if p and p.state != j.state:
            p.save_json()

    def all_points_completed(self):
        for p in self.points:
            if p is not None and p.state not in (STATE_RUNNING, STATE_BROKEN):
                return False
        return True

    def is_sticky_job(self, j):
        """ Return True if the job is the latest running point of a sticky branch. """
        runnings = filter(lambda x: x and x.state == STATE_RUNNING, self.points)
        return self.sticky and runnings and runnings[-1].job_id == j.job_id

    def update_state(self, state):
        """Copy some values taken from the state.runbot file."""
        if self.team_name == 'openerp-dev' and self.name in self.runbot.SERIES_BRANCHES:
            # Don't change stickiness for the hard-coded branches.
            return
        previous = self.sticky
        default_sticky = "stick" in self.name
        self.sticky = state.get(self.team_name, {}).get(self.name, {}).get('sticky', default_sticky)
        if self.sticky != previous:
            log("stickiness changed", team=self.team_name, branch=self.name, now=self.sticky)

    def repo_updates(self):
        return [copy.copy(x) for x in self.all_branches()]

    def all_branches(self):
        r = []
        r.append(self.server)
        r.extend([x for x in self.addons])
        r.append(self.web)
        return r

    def lp_date_last_modified(self):
        return max(p.lp_date_last_modified for p in self.all_branches() if p and p.trigger_build)

    def is_ok(self):
        """
        Test whether the group is useable, i.e. if self.configure_branches()
        didn't early return.
        """
        return self.server and self.web and self.addons

    # TODO all the xxxx_branch argument in __init__ can be moved here.
    def configure_branches(self, launchpad):
        """ Fetch the configured branches. """
        log("runbot-populate-configured-branches")
        self.configured = True
        repo = os.path.join(self.runbot.wd, 'repo')
        filters = {'status': NEW_MERGE_STATUS}

        path = os.path.join(repo, '__configured_' + self.name + '_server')
        b = launchpad.get_branch(unique_name=self.server_branch)
        if not b:
            log("WARNING:no such unique name", name=self.server_branch)
            return
        self.server = RunBotBranch(self.runbot, b, group=self, repo_path=path, project_name='server', trigger_build=True)
        self.server.update_launchpad(b)
        #self.server.merge_count = len(list(b.getMergeProposals(**filters)))

        if self.web_branch:
            path = os.path.join(repo, '__configured_' + self.name + '_web')
            b = launchpad.get_branch(unique_name=self.web_branch)
            if not b:
                log("WARNING:no such unique name", name=self.web_branch)
                return
            self.web = RunBotBranch(self.runbot, b, group=self, repo_path=path, project_name='web', trigger_build=True)
            self.web.update_launchpad(b)
            #self.web.merge_count = len(list(b.getMergeProposals(**filters)))
        else:
            self.web = None

        self.addons = []
        for x in xrange(len(self.addons_branches)):
            path = os.path.join(repo, '__configured_' + self.name + '_addons_' + str(x+1) + 'of' + str(len(self.addons)))
            b = launchpad.get_branch(unique_name=self.addons_branches[x])
            if not b:
                log("WARNING:no such unique name", name=self.addons_branches[x])
                self.addons = []
                return
            bb = RunBotBranch(self.runbot, b, group=self, repo_path=path, project_name='addons', trigger_build=True)
            bb.update_launchpad(b)
            #bb.merge_count = len(list(b.getMergeProposals(**filters)))
            self.addons.append(bb)

    def add_branch(self, b):
        """ Add a single branch (when not using configure_branches(). """
        assert b.name == self.name
        b.group = self
        if b.project_name == 'addons':
            self.addons = [b]
        if b.project_name == 'server':
            self.server = b
        if b.project_name == 'web':
            self.web = b

    # Note: any access to launchpad is done in the same thread.
    def add_main_branches(self):
        """
        Add trunk or 6.{0,1} branches when the complementary branches are missing.
        Also do it when the addons branch is not complete, i.e. when missing
        one of the official addons.
        """
        repo = os.path.join(self.runbot.wd,'repo')

        matched = [s for s in self.runbot.SERIES_BRANCHES if self.name.startswith(s)]
        if matched:
            self.version = matched[0]
            ending = '_' + self.version
        else:
            self.wrong_matching = True
            ending = "_trunk"
            self.version = 'trunk' # TODO this is not accurate

        paths = {
            'addons': os.path.join(repo, 'openerp_openobject-addons_' + self.name_underscore),
            'server': os.path.join(repo, 'openerp_openobject-server_' + self.name_underscore),
            'web': os.path.join(repo, 'openerp_openerp-web_' + self.name_underscore),
        }
        if self.version == '6.0':
            paths['web'] = os.path.join(repo, 'openerp_openobject-client-web_' + self.name_underscore)

        for p in ('server', 'web'):
            if not getattr(self, p):
                b = self.runbot.main_branches[p+ending]
                setattr(self, p, RunBotBranch(self.runbot,b,group=self,repo_path=paths[p]))
                getattr(self, p).update_launchpad(b)
            elif not getattr(self, p).trigger_build:
                b = self.runbot.main_branches[p+ending]
                getattr(self, p).update_launchpad(b)
        if not self.addons:
            b = self.runbot.main_branches['addons'+ending]
            self.addons = [RunBotBranch(self.runbot,b,group=self,repo_path=paths['addons'])]
            self.addons[0].update_launchpad(b)
        elif not self.addons[0].trigger_build:
            b = self.runbot.main_branches['addons'+ending]
            self.addons[0].update_launchpad(b)
    
    def set_reason_ok(self, unique_name, reason_ok, lp_mp_obj=False):
        #unique_name=False set all reason_ok
        for branch in self.all_branches():
            if branch and (branch.unique_name == unique_name or not unique_name):
                branch.reason_ok = reason_ok
                if lp_mp_obj:
                    branch.merge_target = lp_mp_obj.target_branch.unique_name
                    branch.merge_source = lp_mp_obj.source_branch.unique_name
                    branch.merge_web_link = lp_mp_obj.web_link
                    

#----------------------------------------------------------
# OpenERP RunBot Engine
#----------------------------------------------------------

class RunBot(object):
    """Used as a singleton, to manage almost all the Runbot state and logic."""
    def __init__(self, wd, poll, server_net_port, server_xml_port,
        client_web_port, server_longpolling_port, number, nginx_port, domain, test, workers,
        current_job_id, debug, lp, default_job_type='install_all'):
        self.wd=wd
        self.server_net_port=int(server_net_port)
        self.server_xml_port=int(server_xml_port)
        self.client_web_port=int(client_web_port)
        self.server_longpolling_port=int(server_longpolling_port)
        self.number=int(number)
        self.nginx_port=int(nginx_port)
        self.domain=domain
        self.test=int(test)
        self.workers = int(workers)
        self.branches={}
        self.groups={}
        self.main_branches = {}
        self.allocated_port = 0
        self.jobs = {}
        # current_job_id is used to assign build number to groups (thus
        # realizing a queue of groups to be processed).
        self.current_job_id = current_job_id
        self.debug = debug
        self.next_build_number = 0
        self.launchpad = lp
        self.checked_date_last_modified = None
        self.default_job_type = default_job_type
        self.populate_count = 1

    def get_by_host(self, host):
        # Not thread-safe...
        print "Searching for", host
        try:
            d = {}
            for g in self.groups.values():
                if not any(g.points): continue
                for i in g.points:
                    print i
                    if i is not None and \
                        i and (i.job_id > self.current_job_id - self.number or g.is_sticky_job(i)) and \
                        i.running_since and \
                        i.db_prefix == host:
                        print "Found."
                        # Via nginx:
                        d['port'] = str(self.nginx_port)
                        d['name'] = i.db_prefix + '-all'
                        # Directly:
                        d['port'] = str(self.server_xml_port + i.port) # valid for >= 6.1
                        d['name'] = i.db_prefix + '-all'
                        break
                if d: break
            j = simplejson.dumps(d, sort_keys=True, indent=4)
            return j
        except Exception, e:
            print "Exception in get_by_host():", e
            return simplejson.dumps({}, sort_keys=True, indent=4)

    def registered_teams(self):
        return openerprunbot.state.get('registered-teams', []) #+ ['openerp-dev']

    @property
    def SERIES_BRANCHES(self):
        return openerprunbot.state['series']

    @property
    def OLD_SERIES(self):
        return openerprunbot.state['old-series']

    def nginx_reload(self):
        nginx_pid_path = os.path.join(self.wd,'nginx','nginx.pid')
        if os.path.isfile(nginx_pid_path):
            pid=int(open(nginx_pid_path).read())
            try:
                os.kill(pid,1)
            except OSError:
                log('ERROR: cannot reload nginx config')
        else:
            run(["nginx","-p",os.path.abspath(self.wd),"-c",os.path.join(self.wd,"nginx/nginx.conf")])

    def nginx_config(self):
        return openerprunbot.templates.render_template('nginx.conf.mako',r=self)

    def registration_page(self):
        return openerprunbot.templates.render_template('registration.html.mako',r=self)
        
    def state_runbot_page(self):
        return openerprunbot.templates.render_template('state_runbot.html.mako',r=self)

    def nginx_index_time(self,t):
        for m,u in [(86400,'d'),(3600,'h'),(60,'m')]:
            if t>=m:
                return str(int(t/m))+u
        return str(int(t))+"s"

    def nginx_index_sticky(self):
        #branches = [(b.subdomain,b) for b in self.running if b.sticky]
        #branches.sort()
        #return [b for (x,b) in branches]
        return []

    def nginx_groups_sticky(self,team_name):
        gs = [(g.name,g) for g in self.groups.values() if any(g.points) and g.team_name==team_name and g.sticky]
        gs.sort()
        return [g for (x,g) in gs]

    def nginx_groups_others(self,team_name):
        gs = [(g.sticky, g.lp_date_last_modified(), g) for g in self.groups.values() if any(g.points) and g.team_name==team_name and not g.sticky]
        gs.sort(reverse=1)
        gs = [g for (x,y,g) in gs]
        return gs

    def nginx_groups_registered(self,team_name):
        gs = [(g.name,g) for g in self.groups.values() if not any(g.points) and g.team_name==team_name and g.build_number == sys.maxint]
        gs.sort()
        return [g for (x,g) in gs]

    def nginx_index_others(self):
        #branches = [(b.local_date_last_modified,b) for b in self.running if not b.sticky]
        #branches.sort(reverse=1)
        #return [b for (x,b) in branches]
        return []

    def nginx_index(self, team_name):
        #print "nginx_index::team_name", team_name
        try:
            return openerprunbot.templates.render_template('branches.vauxoo.html.mako',
            r=self, t=time.time(), re=re, team_name=team_name, sys=sys, gen_time=time.strftime('%Y-%m-%d %H:%M:%S %Z'))
        except Exception as WTF:
            return openerprunbot.templates.render_template('erros.html.mako', error=WTF)

    def nginx_update(self):
        for team_name in self.teams_with_branches():
            try:
                f = None
                if team_name == 'openerp-dev':
                    fn = 'index.html'
                else:
                    fn = team_name + '.html'
                f = open(os.path.join(self.wd,'static',fn),"w")
                content = self.nginx_index(team_name)
                f.write(content)
            except Exception, e:
                log("WARNING: exception when templating %s:" % fn)
                print traceback.format_exc()
                if f: f.close()
                break

        try:
            f = open(os.path.join(self.wd,'static','register.html'),"w")
            content = self.registration_page()
            f.write(content)
        except Exception, e:
            log("WARNING: exception when templating register.html:")
            print traceback.format_exc()
            if f: f.close()
            
        try:
            f = open(os.path.join(self.wd,'static','state_runbot.html'),"w")
            content = self.state_runbot_page()
            f.write(content)
        except Exception, e:
            log("WARNING: exception when templating state_runbot.html:")
            print traceback.format_exc()
            if f: f.close()

        try:
            f = open(os.path.join(self.wd,'nginx','nginx.conf'),"w")
            content = self.nginx_config()
            f.write(content)
        except Exception, e:
            log("WARNING: exception when templating nginx.conf:")
            print traceback.format_exc()
            if f: f.close()
        self.nginx_reload()

    def teams_with_branches(self):
        ts = [g.team_name for g in self.groups.values()]
        ts = list(set(ts))
        ts.sort()
        return ts

    def allocate_port(self):
        wrap_around = False
        while True:
            if self.allocated_port >= self.number + ADDITIONAL_NUMBER:
                if wrap_around:
                    raise IOError('All available ports seem to be in use, aborting')
                self.allocated_port = 0
                wrap_around = True

            self.allocated_port += 1

            port = self.server_xml_port + self.allocated_port
            # check if port is available, might still be in use
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                s.bind(('0.0.0.0', port))
            except Exception:
                log("WARNING: allocated port still in use, skipping", port=port, allocated_port=self.allocated_port)
                continue
            finally:
                s.close()

            return self.allocated_port

    def process_add(self, b, sticky, team_name):
        log("runbot-process-add", b.unique_name)
        if not project_name_from_unique_name(b.unique_name):
            log("WARNING: can't add branch: project name is not standard.", unique_name=b.unique_name)
            return
        if b.name.startswith(tuple(self.OLD_SERIES)):
            log('WARNING: branch is outdated.', unique_name=b.unique_name)
            return
        if b.unique_name not in self.branches:
            self.branches[b.unique_name] = RunBotBranch(self, b, trigger_build=True)
        bb = self.branches[b.unique_name]
        if (team_name, bb.name) not in self.groups:
            self.groups[(team_name, bb.name)] = RunBotGroupedBranch(self, team_name, bb.name, None, sticky, job_type=self.default_job_type)
        self.groups[(team_name, bb.name)].add_branch(bb)
        bb.update_launchpad(b)
        #filters = {'status': NEW_MERGE_STATUS}
        #try:
        #    bb.merge_count = len(list(b.getMergeProposals(**filters)))
        #except Exception, e:
        #    bb.merge_count = 0

    def register_configured_branches(self):
        
        if eval( open( PDB_SET_TRACE_PATH, "r").read() ).get('"core_py_def_register_configured_branches"', False):
            pdb.set_trace()
        
        for team, v in openerprunbot.state.get('configured-branches', {}).items():
            if not team: continue
            if self.quit: return
            for name, c in v.items():
                if not name: continue
                default_sticky = 'stick' in name
                if c['version'] not in self.SERIES_BRANCHES: continue
                g = RunBotGroupedBranch(self, team, name, c['version'], default_sticky,
                    server_branch=c['server_branch'],
                    client_web_branch=c.get('client_web_branch'),
                    web_branch=c.get('web_branch'),
                    addons_branches=c['addons_branches'],
                    modules=c['modules'])

                # TODO or if it is already there but with different branches.
                if (g.team_name, g.name) not in self.groups:
                    g.configure_branches(self.launchpad)
                    if g.is_ok():
                        log("adding configured group", team=g.team_name, name=g.name)
                        self.groups[(g.team_name, g.name)] = g
                    else:
                        log("WARNING: misconfigured group", team=g.team_name, name=g.name)
                #Start new runbot group in "register branches" without add queue
                #g.build_number = sys.maxint
                #g.sticky = 0
                if not default_sticky:
                    g.need_run_reason = []
        sticky_branch_count = reduce(lambda x,y: x+y, [g.sticky==True and 1 or 0 for g in self.groups.values()])
        if sticky_branch_count >= self.number:
            log("WARNING: self.number reset to new sticky_branch_count+1", sticky_branch_count=sticky_branch_count, old_number=self.number)
            self.number = sticky_branch_count + 1

    # Note: any access to launchpad is done in the same thread.
    def populate_branches(self):
        """Return all LP branches matching our teams and projects."""
        log("runbot-populate-branches")
        # Register main sticky branches
        for s in self.SERIES_BRANCHES:
            parts = {
                'server': '~openerp/openobject-server/' + s,
                'addons': '~openerp/openobject-addons/' + s,
                'web': '~openerp/' + ('openobject-client-web/' if s == '6.0' else 'openerp-web/') + s,
            }
            for p, v in parts.items():
                k = p + '_' + s
                if self.quit: return []
                b = self.launchpad.get_branch(unique_name=v)
                if self.quit: return []
                if not b:
                    log("WARNING: can't get branch:", name=v)
                    continue
                self.process_add(b, 1, 'openerp-dev')
                self.main_branches[k] = b

        # Register other branches
        for v in openerprunbot.state.get('registered-branches', []):
            if self.debug: break
            if self.quit: return []
            m = openerprunbot.branch_input_re.match(v)
            if m:
                if self.quit: return []
                b = self.launchpad.get_branch(unique_name=v)
                if self.quit: return []
                if not b:
                    # TODO better handling of launchpad connection problem
                    log("WARNING: can't get branch:", name=v)
                    continue
                self.process_add(b, 0, m.group(1))
            else:
                log("WARNING: malformed branch name:", v)

        # Register team branches
        for team_name in self.registered_teams():
            if self.debug: break
            if self.quit: return []
            log("Register team branches:", team=team_name)
            team_branches = self.launchpad.get_team_branches(team_name)
            for b in team_branches:
                if project_name_from_unique_name(b.unique_name) and re.search("/(7\.0|trunk|saas-1)-", b.unique_name):
                    print b.unique_name
                    self.process_add(b, 0, team_name)


        # Register configured branches
        self.register_configured_branches()
        
        log("INFO: (MERGE PROPOSAL) self.populate_count %d"%(self.populate_count,) )
        if self.populate_count % 2 == 0:#Check new merge proposals each 2 calls to populate_branches: 4 minutes more time of process.
            log("INFO: (MERGE PROPOSAL) Start Review MP")
            for g in self.groups.values():
                if self.quit: return []
                if not g.configured:
                     g.add_main_branches()
                if 'stick' in g.name:
                    if g.web_branch:
                        log("launchpad update custom sticky web branch " + g.web_branch)
                        b = g.runbot.launchpad.get_branch(unique_name = g.web_branch)
                        if not b:
                            log("WARNING:no such unique name", name = g.web_branch)
                        else:
                            g.web.update_launchpad( b )
                            self.create_rbgb_merge_proposal( g, b, g.web, max_mp=MAX_MP_DEFAULT)
                    if g.server_branch:
                        log("launchpad update custom sticky server branch " + g.server_branch)
                        b = g.runbot.launchpad.get_branch(unique_name = g.server_branch)
                        if not b:
                            log("WARNING:no such unique name", name = g.server_branch)
                        else:
                            g.server.update_launchpad( b )
                            self.create_rbgb_merge_proposal( g, b, g.server, max_mp=MAX_MP_DEFAULT)
                    if g.addons:
                        for addon_rb in g.addons:
                            log("launchpad update custom sticky addons branch " + addon_rb.unique_name)
                            b = g.runbot.launchpad.get_branch(unique_name = addon_rb.unique_name)
                            if not b:
                                log("WARNING:no such unique name", name = addon_rb.unique_name)
                            else:
                                addon_rb.update_launchpad( b )
                                self.create_rbgb_merge_proposal( g, b, addon_rb, max_mp=MAX_MP_DEFAULT)
            log("INFO: (MERGE PROPOSAL) End Review MP")
        self.checked_date_last_modified = max(x.lp_date_last_modified() for x in self.groups.values())
        self.assign_positions()
        self.populate_count += 1
        return self.get_queue()
    
    def create_rbgb_merge_proposal(self, rbgb, lpbranch, rbbranch, max_mp=15):
        filters = {'status': NEW_MERGE_STATUS}
        log("searching merge proposal to " + lpbranch.unique_name)
        merge_proposals = lpbranch.getMergeProposals(**filters)
        if eval( open( PDB_SET_TRACE_PATH, "r").read() ).get('"core_py_def_create_rbgb_merge_proposal"', False):
            pdb.set_trace()
        mp_number__mp_obj = dict([(merge_proposal.web_link.split('/')[-1], merge_proposal) for merge_proposal in merge_proposals])
        mp_numbers_sort = sorted(  mp_number__mp_obj.keys(), reverse=True )
        for mp_number in mp_numbers_sort[:max_mp]:#last MP
            merge_proposal = mp_number__mp_obj[ mp_number ]
            new_branch = merge_proposal.source_branch.unique_name
            new_name = (rbgb.name.replace('sticky', '').replace('stick', '') + '-mp-'+mp_number).replace('--', '-')
            if (rbgb.team_name, new_name) not in self.groups:
                log("Adding merge proposal branch " + new_branch)
                if rbbranch.project_name == 'addons':
                    new_addons_branches = []
                    for old_addon_branch in rbgb.addons_branches:
                        if old_addon_branch == lpbranch.unique_name:
                            new_addons_branches.append( new_branch )
                        else:
                            new_addons_branches.append( old_addon_branch )
                else:
                    new_addons_branches = rbgb.addons_branches
                
                new_modules = ','.join( rbgb.modules )
                if new_modules and rbbranch.project_name == 'addons':
                    if not '!' in new_modules:#"!" is for exclude modules. I dont need exclude modules.
                        #Check mp diff for add new modules
                        mp_diffstat = merge_proposal and merge_proposal.preview_diff and merge_proposal.preview_diff.diffstat or False
                        if mp_diffstat:
                            modules_diff_list = [fname_diff.split('/')[0] for fname_diff in mp_diffstat.keys()]
                            modules_diff_list = list( set( modules_diff_list ) )
                            modules_diff_str = ','.join( modules_diff_list )
                            if modules_diff_str:
                                new_modules += ',' + modules_diff_str
                new_rbgb = RunBotGroupedBranch(self, rbgb.team_name, new_name, rbgb.version, 0,
                    server_branch = rbbranch.project_name == 'server' and new_branch or rbgb.server_branch,
                    client_web_branch = rbbranch.project_name == 'web' and new_branch or rbgb.web_branch,
                    web_branch =  rbbranch.project_name == 'web' and new_branch or rbgb.web_branch,
                    #addons_branches = rbbranch.project_name == 'addons' and new_branch or rbgb.addons_branches,
                    addons_branches = new_addons_branches,
                    modules = new_modules, job_type=rbgb.job_type)
                
                new_rbgb.configure_branches(self.launchpad)
                new_rbgb.set_reason_ok(new_branch, True, merge_proposal)
                
                #Start new runbot group in "register branches" without add queue
                #new_rbgb.build_number = sys.maxint
                #new_rbgb.sticky = 0
                new_rbgb.need_run_reason = []

                if new_rbgb.is_ok():
                    log("adding configured group", team=new_rbgb.team_name, name=new_name)
                    self.groups[(new_rbgb.team_name, new_name)] = new_rbgb
                else:
                    log("WARNING: misconfigured group", team=new_rbgb.team_name, name=new_name)
    

    def assign_positions(self):
        # Sort by last modification date, then assign a position in the queue.
        # Consider only newly modified groups.
        # Don't assign a new position if the position was not yet handled.
        gs = sorted(self.groups.values(), key=lambda x: (-x.sticky, x.lp_date_last_modified()))
        for g in gs:
            if not g.sticky and g.lp_date_last_modified() < self.checked_date_last_modified:
                continue
            if not g.sticky:
                self.checked_date_last_modified = g.lp_date_last_modified()
            if g.need_run_reason and g.build_number == sys.maxint:
                self.next_build_number +=1
                g.build_number = self.next_build_number

    def get_queue(self):
        gs = sorted(self.groups.values(), key=lambda x: x.build_number)
        return filter(lambda g: g.build_number != sys.maxint, gs)

    def available_workers(self):
        return self.workers - len([t for t in threading.enumerate() if t.name.startswith('runbot-group-worker-')])

    def process_command_queue(self):
        while not openerprunbot.queue.empty():
            command, params = openerprunbot.queue.get()
            if command == 'build':
                team_name, group_name = params
                if (team_name, group_name) in self.groups and self.groups[(team_name, group_name)].build_number == sys.maxint:
                    self.next_build_number += 1
                    self.groups[(team_name, group_name)].build_number = self.next_build_number
                    self.groups[(team_name, group_name)].need_run_reason.append('build')
            else:
                log("WARNING: unknown command", command)

    def reset_build_numbers(self):
        gs = [(g.build_number, g) for g in self.groups.values()]
        gs.sort()
        gs = [g for (x,g) in gs]
        self.next_build_number = 0
        sticky_branches = 0
        for g in gs:
            if g.sticky:
                sticky_branches += 1
            if g.build_number == sys.maxint:
                break
            self.next_build_number += 1
            g.build_number = self.next_build_number
        self.number = max(sticky_branches + 1, self.number)

    def complete_jobs(self):
        """Update all slots with the completed jobs."""
        for job in self.jobs.values():
            if job.state in (STATE_RUNNING, STATE_BROKEN):
                for g in self.groups.values():
                    if g.name == job.name:
                        g.complete_point(job)
                        break

    def is_sticky_job(self, job):
        """ Return True if the job is the latest point of a sticky branch. """
        for g in self.groups.values():
            if g.name == job.name and g.team_name == job.team_name:
                return g.is_sticky_job(job)

    def reap_oldest_job(self):
        """Kill some job to limit the number of processes."""
        if len(self.jobs) > self.number:
            victim = None
            for job in self.jobs.itervalues():
                if self.is_sticky_job(job):
                    continue
                if not job.running_since:
                    continue
                if not victim or job.job_id < victim.job_id:
                    victim = job
            if victim:
                victim.stop()
                try:
                    del self.jobs[victim.job_id]
                except KeyError:
                    pass

    def fetch_groups(self):
        """
        Fetch branch information from Launchpad.
        """
        log("runbot-process-grouped-branches")

        self.launchpad.connect()
        gs = self.populate_branches()

        log("runbot-run-group", groups=len(gs))
        return gs

    def run_groups(self, gs):
        for g in gs:
            log("INFO: START def run_groups :: var g", points=str([p and (p.name, p.state) or (None, None) \
                  for p in g.points ]),
                  available_workers=self.available_workers(),
                  jobs=str([(job.name,job.state) for job in self.jobs.values()]),
                  jobs_qty=len(self.jobs),
                  number=self.number
            )            
            if not self.available_workers():
                log("INFO: No process runbot_group because, no available workers", runbot_group_name=g.name)
                break
            if len(self.jobs) > self.number:
                log("INFO: No process runbot_group because, jobs > self.number", runbot_group_name=g.name)
                break
            # Don't run multiple jobs at the same time for the same group.
            if g.need_run_reason and g.all_points_completed():
                port = self.allocate_port()
                self.current_job_id += 1
                job_class = openerprunbot.jobs.JOBS[g.job_type] if g.job_type else openerprunbot.jobs.JOBS[self.default_job_type]
                job = job_class(g, port, self.test, self.current_job_id, self.debug)
                g.add_point(job)
                g.need_run_reason = []
                g.build_number = sys.maxint
                self.jobs[job.job_id] = job
                job.spawn()
            else:
               if not g.need_run_reason:
                  log("INFO: No process runbot_group because need_run_reason is not set", runbot_group_name=g.name, runbot_group_name_need_run_reason=g.need_run_reason)
               if not g.all_points_completed():
                  log("INFO: No process runbot_group because not all_points_completed()", runbot_group_name=g.name, g_all_points_completed=g.all_points_completed() )


    def loop(self):
        """
        Repeatedly fetch branch information from Launchpad, run jobs, update
        HTML pages, ...
        """
        self.quit = False
        def signal_handler(sig, frame):
            log("SIGINT or SIGTERM received, exiting...")
            openerprunbot.server.stop_server()
            for i in self.jobs.values():
                i.stop()
            self.quit = True
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        while not self.quit:
            if eval( open( PDB_SET_TRACE_PATH, "r").read() ).get('core_py_loop', False):
                pdb.set_trace()
            gs = []
            try:
                gs = self.fetch_groups()
            except Exception, e:
                log("WARNING: exception:")
                log(traceback.format_exc())
            
            for i in xrange(12): # 2 minutes
                if self.quit: break
                try: # Make sure time.sleep() is always called.
                    self.process_command_queue()
                    self.run_groups(gs)
                    self.complete_jobs()
                    self.reap_oldest_job()
                    self.reset_build_numbers()
                    self.nginx_update()
                    for g in self.groups.values():
                        g.update_state(openerprunbot.state)
                except Exception, e:
                    log("WARNING: exception:")
                    log(traceback.format_exc())
                time.sleep(10)

    def read_state(self):
        with runbot_state_lock:
            try:
                path = os.path.join('static', 'state.runbot')
                if os.path.exists(path):
                    with open(path, 'r') as f:
                        state = simplejson.loads(f.read())
                        #openerprunbot.state = state
                        #return openerprunbot.state
                        data = simplejson.dumps(state, sort_keys=True, indent=2)
                        return data
            except Exception, e:
                return "Exception in read_state(): ", e
            
