import os
import shutil
import signal
import subprocess
import threading
import time
import traceback
import pdb

from ..misc import *

# Job states
STATE_ALLOCATED = 'allocated'
STATE_BROKEN = 'broken'
STATE_PULLING = 'pulling'
STATE_RUNNING = 'running'
STATE_TESTING = 'testing'

JOB7_PATH = os.path.realpath(os.path.join(os.path.dirname(__file__), '..', '..'))
PDB_SET_TRACE_PATH = os.path.join( os.path.normpath( os.path.dirname( __file__ ) ), '..', '..', "pdb_set_trace")

class InstallAllJob(object):
    """
    A Job encapsulates all the necessary data to build a branch group for a
    given slot. The build is done in its own worker thread.
    """

    def __init__(self, g, port, test, job_id, debug):
        self.job_id = job_id
        self.state = STATE_ALLOCATED
        self.team_name = g.team_name
        self.name = g.name
        self.name_underscore = g.name_underscore
        self.name_dashes = g.name_dashes
        self.version = g.version
        self.debug = debug
        self.need_run_reason = g.need_run_reason

        self.repo_updates = g.repo_updates()
        self.port = port
        self.test = test
        self.running_server_pid = 0
        self.client_web_pid = 0
        self.running_since=0

        repo = os.path.join(g.runbot.wd,'repo')
        job_suffix = '_job-' + str(self.job_id)
        self.server_src = g.server.repo_path + job_suffix
        self.client_web_src = (g.web.repo_path + job_suffix) if g.version == '6.0' and g.web else None
        self.web_src = (g.web.repo_path + job_suffix) if g.web else None

        # if addons is not the full addons branch use trunk
        self.addons_src = [a.repo_path + job_suffix for a in g.addons]

        # Running path <Root>/static/<domain>
        self.subdomain = "%s-%s-%s"%(self.team_name, self.name.replace('_','-').replace('.','-'),self.job_id)
        self.running_path = os.path.join(g.runbot.wd,'static',self.subdomain)
        self.json_path = g.json_path
        self.log_path = os.path.join(self.running_path,'logs')
        self.flags_path = os.path.join(self.running_path,'flags')
        self.tmp_path = os.path.join(self.running_path,'tmp')

        # Server
        self.server_path=os.path.join(self.running_path,"server")
        self.server_bin_path=os.path.join(self.server_path,"openerp-server")
        self.server_log_path=os.path.join(self.log_path,'server.txt')
        self.server_log_base_path=os.path.join(self.log_path,'test-base.txt')
        self.server_log_all_path=os.path.join(self.log_path,'test-all.txt')

        # coverage
        self.coverage_file_path=os.path.join(self.log_path,'coverage.pickle')
        self.coverage_base_path=os.path.join(self.log_path,'coverage-base')
        self.coverage_all_path=os.path.join(self.log_path,'coverage-all')

        # Web60
        self.client_web_pid=None
        self.client_web_path=os.path.join(self.running_path,"client-web")
        self.client_web_bin_path=os.path.join(self.client_web_path,"openerp-web.py")
        self.client_web_doc_path=os.path.join(self.client_web_path,"doc")
        self.client_web_log_path=os.path.join(self.log_path,'client-web.txt')

        # test
        self.test_base_path=os.path.join(self.log_path,'test-base.txt')
        self.test_result=None
        self.test_all_path=os.path.join(self.log_path,'test-all.txt')

        self.server_net_port=g.runbot.server_net_port
        self.server_xml_port=g.runbot.server_xml_port
        self.client_web_port=g.runbot.client_web_port
        self.server_longpolling_port=g.runbot.server_longpolling_port

        if self.version == '6.0':
            self.db_prefix = '%s_%s' % (self.name_underscore, self.job_id)
            self.db = '%s_%s_base' % (self.name_underscore, self.job_id)
            self.db_all = '%s_%s_all' % (self.name_underscore, self.job_id)
        else:
            self.db_prefix = '%s-%s' % (self.name_dashes, self.job_id)
            self.db = '%s-%s-base' % (self.name_dashes, self.job_id)
            self.db_all = '%s-%s-all' % (self.name_dashes, self.job_id)

        for p in ('addons', 'server', 'web'):
            setattr(self, p + '_committer_name', None)
            setattr(self, p + '_committer_xgram', None)
            setattr(self, p + '_committer_email', None)
        self.start_time = 0
        self.completed_time = 0

        self.modules = g.modules


        self.server_rev = g.server.local_revision_count
        self.addons_rev = [a.local_revision_count for a in g.addons]
        self.web_rev = g.web.local_revision_count if g.web else 0
        self.build_number = g.build_number

    def spawn(self):
        log("runbot-spawn-worker-" + str(self.job_id), group=self.name)
        t = threading.Thread(target=self.work, name=('runbot-group-worker-' + str(self.job_id)))
        t.daemon = True
        t.start()

    def work(self):
        if eval( open(PDB_SET_TRACE_PATH, "r").read() ).get("install_all_def_work", False):
            pdb.set_trace()
        try:
            log("start new work")
            self.state = STATE_PULLING
            log("self.state" + repr(self.state))
            r = self.pull_branches()
            log("pull_branches" + repr(r))
            if r:
                #self.state = STATE_TESTING#Here no start to test, only start rsync
                log("self.state" + repr( self.state) )
                self.start_time = time.time()
                self.start()
                self.completed_time = time.time()
                self.state = STATE_RUNNING
                log("runbot-end-worker", job=self.name)
            else:
                self.state = STATE_BROKEN
                log("runbot-end-worker [failed pull]", job=self.name)
        except Exception, e:
            self.state = STATE_BROKEN
            log("runbot-end-worker [with exception]", job=self.name)
            log( repr(traceback.format_exc()) )

    def pull_branches(self):
        job_suffix = '_job-' + str(self.job_id)
        for b in self.repo_updates:
            log("branch-update",branch=b.unique_name)
            repo_path = b.repo_path + job_suffix
            if os.path.exists(repo_path):
                log("bzr pull " + repr(["bzr", "pull", "lp:%s" % b.unique_name, "--quiet", "--remember", "--overwrite", "-d", repo_path]) )
                rc = run(["bzr", "pull", "lp:%s" % b.unique_name, "--quiet", "--remember", "--overwrite", "-d", repo_path])
                if rc: return False
            else:
                log("bzr branch " + repr( ["bzr", "branch", "--quiet", "--no-tree", "lp:%s" % b.unique_name, repo_path ] ) )
                rc = run(["bzr", "branch", "--quiet", "--no-tree", "lp:%s" % b.unique_name, repo_path ])
                if rc: return False
#            rc = run(["bzr", "update", "-r", str(b.local_revision_count), repo_path, "--quiet"])
# to be done a bzr export
            if rc: return False
            committer_name, committer_xgram, committer_email = \
                get_committer_info(repo_path)
            b.committer_name = committer_name
            b.committer_xgram = committer_xgram
            b.committer_email = committer_email
            log("get-commiter-info", name=committer_name, xgram=committer_xgram, email=committer_email)
            
            log("get-revno-info "+ b.unique_name)
            revno = get_revno_info( repo_path )
            if revno and b.local_revision_count != revno:
                log("set-revno-info "+ b.unique_name + " - current revno:" + repr(b.local_revision_count) + " - new revno: " + repr( revno ) )
                b.local_revision_count = revno
            else:
                log("set-revno-info "+ b.unique_name + " no changes of revno.")
            """
            if b.merge_target:
                #Apply merge proposal to developer branch
                cmd = ["bzr", "reconfigure", "--tree", repo_path]
                log( repr( cmd ) )
                rc = run(cmd)
                log("rc_return: " + repr( rc ) )
                
                cmd = ["bzr", "merge", "lp:%s" % b.merge_target, "-d", repo_path]
                log( repr( cmd ) )
                rc = run(cmd)
                log("rc_return: " + repr( rc ) )
                
                cmd = ["bzr", "resolve", "--all"]
                log( repr( cmd ) )
                rc = run(cmd)
                log("rc_return: " + repr( rc ) )
                
                cmd = ["bzr", "commit", "-m", "[MERGE] Auto merge from runbot", repo_path]
                log( repr( cmd ) )
                rc = run(cmd)
                log("rc_return: " + repr( rc ) )
                
                cmd = ["bzr", "reconfigure", "--branch", repo_path]
                log( repr( cmd ) )
                rc = run(cmd)
                log("rc_return: " + repr( rc ) )
            """
        return True

    def export_and_move(self, branch, tmp, dest, chroot="."):
        if os.path.isdir(tmp):
            shutil.rmtree(tmp)
        log("bzr export " + repr( ["bzr","export", self.tmp_path, branch] ))
        run(["bzr","export", self.tmp_path, branch])
        root = os.path.join(self.tmp_path, chroot)
        for i in os.listdir(root):
            src = os.path.join(root, i)
            dst = os.path.join(dest, i)
            if os.path.isdir(dst):
                shutil.rmtree(dst)
            elif os.path.isfile(dst):
                os.unlink(dst)
            os.rename(src,dst)

    def start_rsync(self):
        if eval( open( PDB_SET_TRACE_PATH, "r").read() ).get("install_all_def_start_rsync", False):
            pdb.set_trace()
        log("job-start-rsync",branch=self.name)
        for i in [self.server_path, self.log_path, self.flags_path, self.client_web_doc_path]:
            if not os.path.exists(i):
                os.makedirs(i)
        # copy server
        self.export_and_move(self.server_src, self.tmp_path, self.server_path)

        # copmute addons path (6.0 uses bin/addons, 6.1, 7.0 and trunk use openerp/addons)
        addons_dest = "openerp/addons"
        if not os.path.exists(os.path.join(self.server_path,addons_dest)):
            addons_dest = "bin/addons"
        addons_path = os.path.join(self.server_path,addons_dest)

        for a in self.addons_src:
            self.export_and_move(a, self.tmp_path, addons_path)

        if self.web_src and self.version != '6.0':
            self.export_and_move(self.web_src, self.tmp_path, addons_path, "addons")

        # @vmt wtf ?
        if os.path.exists(os.path.join(self.server_path,"openerp/tests/addons")) and addons_dest == "openerp/addons":
            run(["rsync","-a",os.path.join(self.server_path,"openerp/tests/addons/"), addons_path])

        # copy web-client
        if self.client_web_src:
            self.export_and_move(self.client_web_src, self.tmp_path, self.client_web_path)
        log("job-grep",branch=self.name)
            
        current_path = os.path.join( os.getcwd(), 'static' ) + os.sep
        cmd = "grep -rl a2aevl8w " + current_path + "  | xargs sed -i 's/a2aevl8w/admin/g'"
        log("start grep",cmd=cmd)
        subprocess.call(cmd, shell=True)
        log("end grep",cmd=cmd)
        
        current_path = os.path.join( os.getcwd(), 'repo' ) + os.sep
        cmd = "grep -rl a2aevl8w " + current_path + "  | xargs sed -i 's/a2aevl8w/admin/g'"
        log("start grep",cmd=cmd)
        subprocess.call(cmd, shell=True)
        log("end grep",cmd=cmd)


    def start_createdb(self, no_create=False):
        log("start pg_terminate_backend")
        run(["psql","postgres","-c","select pg_terminate_backend(pid) from pg_stat_activity where datname in ('%s','%s')"%(self.db,self.db_all)])
        time.sleep(3)
        log("start dropdb")
        run(["dropdb",self.db])
        run(["dropdb",self.db_all])
        if not no_create:
            log("createdb base")
            run(["createdb",self.db])
            log("createdb all")
            run(["createdb",self.db_all])

    def run_log(self, cmd, logfile, env=None):
        env = dict(os.environ, **env) if env else None
        log("run", *cmd, logfile=logfile)
        out=open(logfile,"w")
        p=subprocess.Popen(cmd, stdout=out, stderr=out, close_fds=True, env=env)
        self.running_server_pid=p.pid
        p.communicate()
        return p

    def resolve_server_bin_path(self):
        # This can be done only if the files are present otherwise the if
        # will always fail. Alternatively, server_bin_path could be a property.
        if not os.path.exists(self.server_bin_path): # for 6.0 branches
            self.server_bin_path=os.path.join(self.server_path,"bin","openerp-server.py")

    def start_test_base(self):
        log("job-start-server-base")
        cmd = [self.server_bin_path,"-d",self.db,"-i","base","--stop-after-init","--no-xmlrpc","--no-xmlrpcs","--log-level=test"]
        _has_test_enable_flag = False
        if has_test_enable_flag(self.server_bin_path):
            cmd.append("--test-enable")
            _has_test_enable_flag = True
        if has_no_netrpc_flag(self.server_bin_path):
            cmd.append("--no-netrpc")
        #cmd = ["coverage","run","--branch"] + cmd
        self.run_log(cmd, logfile=self.test_base_path,env={'COVERAGE_FILE': self.coverage_file_path})
        #run(["coverage","html","-d",self.coverage_base_path,"--ignore-errors","--include=*.py"],env={'COVERAGE_FILE': self.coverage_file_path})
        if _has_test_enable_flag:
            success_message = "openerp.modules.loading: Modules loaded."
            rc = not bool(run(["grep",success_message,self.test_base_path]))
        else:
            rc = bool(run(["grep","Traceback",self.test_base_path]))
        return rc

    def start_test_all(self, without_demo=False):
        if eval( open( PDB_SET_TRACE_PATH, "r").read() ).get("install_all_def_start_test_all", False):
            pdb.set_trace()
        log("job-start-server-all")
        mods = []
        if not os.path.exists(os.path.join(self.server_path,"openerp/addons")):
            mods = os.listdir(os.path.join(self.server_path,"bin/addons"))
        elif os.path.exists(os.path.join(self.server_path,"openerp/addons")):
            mods = os.listdir(os.path.join(self.server_path,"openerp/addons"))
        if os.path.exists(os.path.join(self.server_path,"openerp/tests/addons")):
            mods += os.listdir(os.path.join(self.server_path,"openerp/tests/addons"))
        bl = [
            'document_ftp', 'l10n_lu', 'crm_partner_assign', 'stock_location', #stock_location no generate a procurement :(
            '.bzrignore', '__init__.pyc', '__init__.py', 'base_quality_interrogation.py',
        ]
        mods = [m for m in mods if m not in bl]
        if self.modules:
            # TODO feedback when a requested modules does'nt exist.
            mods = [m for m in mods if m in self.modules]
        mods = ",".join(mods)
        cmd = [self.server_bin_path,"-d",self.db_all,"-i",mods,"--stop-after-init","--no-xmlrpc","--no-xmlrpcs","--log-level=test"]
        _has_test_enable_flag = False
        if has_test_enable_flag(self.server_bin_path):
            cmd.append("--test-enable")
            _has_test_enable_flag = True
        if has_no_netrpc_flag(self.server_bin_path):
            cmd.append("--no-netrpc")
        if without_demo:
            cmd.append("--without-demo=True")
        #cmd = ["coverage","run","--branch"] + cmd
        self.run_log(cmd, logfile=self.test_all_path)
        #run(["coverage","html","-d",self.coverage_all_path,"--ignore-errors","--include=*.py"])
        if _has_test_enable_flag:
            success_message = "openerp.modules.loading: Modules loaded."
            rc = not bool(run(["grep",success_message,self.test_all_path]))
        else:
            rc = bool(run(["grep","Traceback",self.test_all_path]))
        return rc

    def start_server(self):
        port = self.port
        log("job-start-server",branch=self.name,port=port)
        cmd=[self.server_bin_path,"--no-xmlrpcs","--xmlrpc-port=%d"%(self.server_xml_port+port)]
        if has_no_netrpc_flag(self.server_bin_path):
            cmd.append("--netrpc-port=%d" % (self.server_net_port+port))
        if os.path.exists(os.path.join(self.server_path, 'openerp')):
            cmd+=["--db-filter=%d(-.*)?$","--load=web"]
        if os.path.exists(os.path.join(self.server_path, 'openerp', 'addons', 'im')):
            cmd += ["--workers", "1"]
            cmd += ["--max-cron-threads", "1"]
            cmd += ["--longpolling-port", str(self.server_longpolling_port + port)]
            # Preload the registries, but more importantly, make sure only
            # those two databases are loaded by the cron worker.
            cmd += ["-d", "%s,%s" % (self.db, self.db_all)]
        log("run",*cmd,log=self.server_log_path)
        out=open(self.server_log_path,"w")
        p=subprocess.Popen(cmd, stdout=out, stderr=out, close_fds=True)
        self.running_server_pid=p.pid

    def start_client_web(self):
        if not self.client_web_src:
            return
        port = self.port
        log("job-start-client-web",branch=self.name,port=port)
        config="""
        [global]
        server.environment = "development"
        server.socket_host = "0.0.0.0"
        server.socket_port = %d
        server.thread_pool = 10
        tools.sessions.on = True
        log.access_level = "INFO"
        log.error_level = "INFO"
        tools.csrf.on = False
        tools.log_tracebacks.on = False
        tools.cgitb.on = True
        openerp.server.host = 'localhost'
        openerp.server.port = '%d'
        openerp.server.protocol = 'socket'
        openerp.server.timeout = 450
        [openerp-web]
        dblist.filter = 'BOTH'
        dbbutton.visible = True
        company.url = ''
        openerp.server.host = 'localhost'
        openerp.server.port = '%d'
        openerp.server.protocol = 'socket'
        openerp.server.timeout = 450
        """%(self.client_web_port+port,self.server_net_port+port,self.server_net_port+port)
        config=config.replace("\n        ","\n")
        cfgs = [os.path.join(self.client_web_path,"doc","openerp-web.cfg"), os.path.join(self.client_web_path,"openerp-web.cfg")]
        for i in cfgs:
            f=open(i,"w")
            f.write(config)
            f.close()

        cmd=[self.client_web_bin_path]
        log("run",*cmd,log=self.client_web_log_path)
        out=open(self.client_web_log_path,"w")
        p=subprocess.Popen(cmd, stdout=out, stderr=out, close_fds=True)
        self.client_web_pid=p.pid

    def start(self):
        if eval( open( PDB_SET_TRACE_PATH, "r").read() ).get('"install_all_def_start"', False):
            pdb.set_trace()
        log("job-start",branch=self.name,port=self.port)
        self.start_rsync()
        log("end: start_rsync")
        log("start: resolve_server_bin_path")
        self.resolve_server_bin_path()
        log("end: resolve_server_bin_path")
        self.state = STATE_TESTING#Here rsync ended and now start to test
        try:
            if self.test:
                if self.version and self.version.startswith('6.'):
                    self.start_createdb()
                    r = self.start_test_base()
                    if not self.debug:
                        r = r and self.start_test_all()
                    self.test_result = 'ok' if r else 'error'
                else:
                    self.start_createdb(no_create=True)
                    nonetrpc = has_no_netrpc_flag(self.server_bin_path)
                    self.test_result = test_gte_7(self, no_no_netrpc=not nonetrpc)

            self.start_server()
            self.start_client_web()
        except OSError,e:
            log("branch-start-error",exception=e)
        except IOError,e:
            log("branch-start-error",exception=e)
        self.running_since=time.time()
        log("branch-started",branch=self.name,port=self.port)

    def stop(self):
        if eval( open( PDB_SET_TRACE_PATH, "r").read() ).get('"install_all_def_stop"', False):
            pdb.set_trace()
        log("Stopping job", id=self.job_id, branch=self.name)
        if self.running_server_pid:
            kill(self.running_server_pid, signal.SIGTERM)
            threading.Timer(3, kill, [self.running_server_pid, signal.SIGKILL]).start()
            self.running_server_pid = 0
        if self.client_web_pid:
            kill(self.client_web_pid)
            threading.Timer(3, kill, [self.client_web_pid, signal.SIGKILL]).start()
            self.client_web_pid = 0
        #ToDo: Validate psql --version
        subprocess.call(["psql","postgres","-c","select pg_terminate_backend(pid) from pg_stat_activity where datname in ('%s','%s')"%(self.db,self.db_all)])#Postgres < 9.2
        #subprocess.call(["psql","postgres","-c","select pg_terminate_backend(procpid) from pg_stat_activity where datname in ('%s','%s')"%(self.db,self.db_all)])#Postgres >= 9.2
        
        subprocess.call(["psql","postgres","-c",'DROP DATABASE IF EXISTS "%s";'%(self.db)])#Standard delete
        subprocess.call(["psql","postgres","-c",'DROP DATABASE IF EXISTS "%s";'%(self.db_all)])#Standard delete
        run(["rm", "-rf", self.server_path])

    def save_json(self):
        """Append the point data to a JSON file for posterity."""
        # A job must be saved only once.
        if hasattr(self, 'json_saved'):
            log("=== save_json() called more than once. ===", job_id=self.job_id)
        self.json_saved = True

        path = self.json_path
        state = {}

        if os.path.exists(path):
            with open(path, 'r') as h:
                state = simplejson.loads(h.read())

        value = {}
        committers = []
        for p in ('addons', 'server', 'web'):
            committers += [
                p + '_committer_name',
                p + '_committer_xgram',
                # p + '_committer_email',
            ]
        for a in [
            'job_id', 'db', 'running_path', 'subdomain', 'server_rev',
            'addons_rev', 'web_rev', 'need_run_reason',
            'test_result',
            ] + committers:
            value[a] = getattr(self, a)
        state.setdefault('jobs', [])
        state['jobs'].append(value)

        with open(path, 'w') as h:
            data = simplejson.dumps(state, sort_keys=True, indent=4)
            h.write(data)

def test_gte_7(job, no_no_netrpc=False):
    """test version >= 7.0"""
    db_prefix = '%s-%s' % (job.name_dashes, job.job_id)
    values = {
        'BINPATH': JOB7_PATH,
        'db_prefix': db_prefix,
        'log_prefix': job.log_path,
        'server': job.server_path,
        'addons': os.path.join(job.server_path, "openerp/addons"),
        'web': os.path.join(job.server_path, "openerp"),    # TODO make it optional
        'port': str(job.server_xml_port + job.port),
    }
    command = """
    %(BINPATH)s/runbot-job-7
    --db-prefix %(db_prefix)s
    --log-prefix %(log_prefix)s
    --server-path %(server)s
    --addons-path %(addons)s
    --web-path %(web)s
    --test-port %(port)s
    """ % values
    if job.modules:
         command += " --modules %s "%( ','.join( job.modules ) )
    command = command.split()
    if no_no_netrpc:
        command.append('--no-no-netrpc')
    r = run(command)

    return {
        0: 'ok',
        1: 'warning',
        2: 'error',
        3: 'killed',
        4: 'broken',
    }.get(r, 'other')
